from permissions_auditor.processors import auth_mixins
from permissions_auditor.tests.base import ProcessorTestCase
from permissions_auditor.tests.fixtures import views


class MixinProcessorTestCaseMixin:
    """Mixins should never be able to process function based views."""

    def assert_cannot_process_non_cbvs(self):
        self.assertCannotProcess([
            views.base_view, views.BaseView,
            views.login_required_view,
            views.staff_member_required_view,
            views.active_user_required_view,
            views.anonymous_user_required_view,
            views.superuser_required_view,
            views.user_passes_test_view
        ])


class TestLoginRequiredMixinProcessor(MixinProcessorTestCaseMixin, ProcessorTestCase):

    def setUp(self):
        self.processor = auth_mixins.LoginRequiredMixinProcessor()
        self.expected_results = {'permissions': [], 'login_required': True, 'docstring': None}

    def test_cannot_process(self):
        self.assertCannotProcess([
            views.PermissionRequiredView, views.PermissionRequiredMultiView,
            views.PermissionRequiredViewNoPerm,
            views.PermissionRequiredViewDocstring, views.PermissionRequiredViewNoDocstring,
            views.UserPassesTestView, views.UserPassesTestViewCustomFunc,
            views.UserPassesTestViewDocstring, views.UserPassesTestViewNoDocstring
        ])

    def test_cb_loginrequiredview(self):
        self.assertCanProcessView(views.LoginRequiredView, **self.expected_results)

    def test_cb_inheritedloginrequiredview(self):
        self.assertCanProcessView(views.InheritedLoginRequiredView, **self.expected_results)


class TestPermissionRequiredMixinProcessor(MixinProcessorTestCaseMixin, ProcessorTestCase):

    def setUp(self):
        self.processor = auth_mixins.PermissionRequiredMixinProcessor()

    def test_cannot_process(self):
        self.assertCannotProcess([
            views.LoginRequiredView,
            views.UserPassesTestView, views.UserPassesTestViewCustomFunc,
            views.UserPassesTestViewDocstring, views.UserPassesTestViewNoDocstring
        ])

    def test_cb_permissionsrequiredview(self):
        self.assertCanProcessView(
            views.PermissionRequiredView,
            permissions=['tests.test_perm'], login_required=True, docstring=None
        )

    def test_cb_inherit_permissionsrequiredview(self):
        self.assertCanProcessView(
            views.InheritedPermissionRequiredView,
            permissions=['tests.test_perm'], login_required=True, docstring=None
        )

    def test_cb_permissionsrequiredview_no_perm(self):
        """
        Views that override has_permission() and do not set permission_required should be processed.
        """
        self.assertCanProcessView(
            views.PermissionRequiredViewNoPerm,
            permissions=[], login_required=True, docstring='The user\'s first name must be Bob'
        )

    def test_cb_permissionsrequiredview_docstring(self):
        """Views that implement has_permission() and have a docstring should be retrieved."""
        self.assertCanProcessView(
            views.PermissionRequiredViewDocstring,
            permissions=['tests.test_perm'], login_required=True,
            docstring='Custom docstrings should be detected.'
        )

    def test_cb_permissionsrequiredview_no_docstring(self):
        """
        Views that implement has_permission() and do not have a docstring
        should return a default messsage.
        """
        self.assertCanProcessView(
            views.PermissionRequiredViewNoDocstring,
            permissions=['tests.test_perm'], login_required=True,
            docstring='Custom (no docstring found)'
        )

    def test_cb_permissionrequiredview_multi(self):
        """
        Views with multiple permissions should return all permissions.
        """
        self.assertCanProcessView(
            views.PermissionRequiredMultiView,
            permissions=['tests.test_perm', 'tests.test_perm2'], login_required=True, docstring=None
        )


class TestUserPassesTestMixinProcessor(MixinProcessorTestCaseMixin, ProcessorTestCase):

    def setUp(self):
        self.processor = auth_mixins.UserPassesTestMixinProcessor()

    def test_cannot_process(self):
        self.assertCannotProcess([
            views.LoginRequiredView,
            views.PermissionRequiredView, views.PermissionRequiredMultiView,
            views.PermissionRequiredViewNoPerm,
            views.PermissionRequiredViewDocstring, views.PermissionRequiredViewNoDocstring
        ])

    def test_cb_userpassestestview(self):
        self.assertCanProcessView(
            views.UserPassesTestView,
            permissions=[], login_required=None, docstring='Custom (no docstring found)'
        )

    def test_cb_inheriteduserpassestestview(self):
        self.assertCanProcessView(
            views.InheritedUserPassesTestView,
            permissions=[], login_required=None, docstring='Custom (no docstring found)'
        )

    def test_cb_userpassestestview_docstring(self):
        """Views that implement test_func() and have a docstring should be retrieved."""
        self.assertCanProcessView(
            views.UserPassesTestViewDocstring,
            permissions=[], login_required=None, docstring='Custom docstrings should be detected.'
        )

    def test_cb_userpassestestview_no_docstring(self):
        """
        Views that implement test_func() and do not have a docstring
        should return a default messsage.
        """
        self.assertCanProcessView(
            views.UserPassesTestViewNoDocstring,
            permissions=[], login_required=None, docstring='Custom (no docstring found)'
        )

    def test_cb_userpassestestview_custom_func(self):
        """
        Views that override get_test_func() should check the new function returned
        instead of the default test_func() function.
        """
        self.assertCanProcessView(
            views.UserPassesTestViewCustomFunc,
            permissions=[], login_required=None, docstring='Custom docstrings should be detected.'
        )



from io import StringIO
from django.core.management import call_command
from django.test import TestCase


class CheckViewPermsTest(TestCase):

    def test_missing_perms(self):
        out = StringIO()
        call_command('check_view_permissions', stdout=out)
        self.assertIn('`tests.test_perm`', out.getvalue())
        self.assertIn('`tests.test_perm2`', out.getvalue())


class CheckDumpViewPermsTest(TestCase):

    def test_dump_perms_no_args(self):
        out = StringIO()
        call_command('dump_view_permissions', stdout=out)

        # Default output should be JSON formatted
        self.assertIn('[{"module": "permissions_auditor.tests.fixtures.views"', out.getvalue())

    def test_dump_perms_json(self):
        out = StringIO()
        call_command('dump_view_permissions', format='json', stdout=out)

        # Output should be JSON formatted
        self.assertIn('[{"module": "permissions_auditor.tests.fixtures.views"', out.getvalue())

    def test_dump_perms_csv(self):
        out = StringIO()
        call_command('dump_view_permissions', format='csv', stdout=out)

        # CSV output header should be present
        self.assertIn('module,name,url,permissions,login_required,docstring', out.getvalue())



"""Processors for django auth decorators."""

import inspect

from .base import BaseDecoratorProcessor


class PermissionRequiredDecoratorProcessor(BaseDecoratorProcessor):
    """
    Process ``@permission_required()`` decorator.
    """

    def can_process(self, view):
        if inspect.isclass(view):
            for func in self._get_method_decorators(view.dispatch):
                if 'user_passes_test' not in (func.__name__, func.__qualname__.split('.')[0]):
                    continue

                test_func = inspect.getclosurevars(func).nonlocals['test_func']
                if test_func.__name__ == 'check_perms':
                    return True

        elif inspect.isfunction(view):
            # Unwrap the function and look for the has_perms property.
            return self._has_func_decorator(view, 'has_perms')

        return False

    def get_permission_required(self, view):
        permissions = []

        if inspect.isclass(view):
            for func in self._get_method_decorators(view.dispatch):
                if 'user_passes_test' not in (func.__name__, func.__qualname__.split('.')[0]):
                    continue

                test_func = inspect.getclosurevars(func).nonlocals['test_func']
                if test_func.__name__ == 'check_perms':
                    closures = inspect.getclosurevars(test_func).nonlocals
                    if 'perm' in closures:
                        perm = closures['perm']

                        # Ensure perm is not a function
                        if not inspect.isfunction(perm):
                            if isinstance(perm, str):
                                permissions.append(perm)
                            else:
                                permissions.extend(perm)

        elif inspect.isfunction(view) and self._has_test_func(view):
            for closure in self._get_test_func_closures(view):
                if 'perm' in closure.nonlocals:
                    perm = closure.nonlocals['perm']

                    # Ensure perm is not a function
                    if not inspect.isfunction(perm):
                        if isinstance(perm, str):
                            permissions.append(perm)
                        else:
                            permissions.extend(perm)

        return permissions

    def get_login_required(self, view):
        return True


class LoginRequiredDecoratorProcessor(BaseDecoratorProcessor):
    """
    Process ``@login_required`` decorator.
    """

    def can_process(self, view):
        if inspect.isclass(view):
            return self._has_method_decorator(view.dispatch, 'login_required')

        elif inspect.isfunction(view):
            # Unwrap the function and look for the is_authenticated property.
            return self._has_func_decorator(view, 'is_authenticated')

        return False

    def get_login_required(self, view):
        return True


class StaffMemberRequiredDecoratorProcessor(BaseDecoratorProcessor):
    """
    Process Django admin's ``@staff_member_required`` decorator.
    """

    def can_process(self, view):
        if inspect.isclass(view):
            return self._has_method_decorator(view.dispatch, 'staff_member_required')

        elif inspect.isfunction(view):
            # Unwrap the function and look for the is_staff property.
            return self._has_func_decorator(view, 'is_staff')

        return False

    def get_login_required(self, view):
        return True

    def get_docstring(self, view):
        return 'Staff member required'


class ActiveUserRequiredDecoratorProcessor(BaseDecoratorProcessor):
    """
    Process ``@user_passes_test(lambda u: u.is_active)`` decorator.
    """

    def can_process(self, view):
        if inspect.isclass(view) and self._has_method_decorator(view.dispatch, 'user_passes_test'):
            return self._has_test_func_lambda(view.dispatch, 'is_active')

        elif inspect.isfunction(view):
            # Unwrap the function and look for the is_active property.
            return self._has_func_decorator(view, 'is_active')

        return False

    def get_login_required(self, view):
        return True

    def get_docstring(self, view):
        return 'Active user required'


class AnonymousUserRequiredDecoratorProcessor(BaseDecoratorProcessor):
    """
    Process ``@user_passes_test(lambda u: u.is_anonymous)`` decorator.
    """

    def can_process(self, view):
        if inspect.isclass(view) and self._has_method_decorator(view.dispatch, 'user_passes_test'):
            return self._has_test_func_lambda(view.dispatch, 'is_anonymous')

        elif inspect.isfunction(view):
            # Unwrap the function and look for the is_anonymous property.
            return self._has_func_decorator(view, 'is_anonymous')

        return False

    def get_docstring(self, view):
        return 'Anonymous user required'


class SuperUserRequiredDecoratorProcessor(BaseDecoratorProcessor):
    """
    Process ``@user_passes_test(lambda u: u.is_superuser)`` decorator.
    """

    def can_process(self, view):
        if inspect.isclass(view) and self._has_method_decorator(view.dispatch, 'user_passes_test'):
            return self._has_test_func_lambda(view.dispatch, 'is_superuser')

        elif inspect.isfunction(view):
            # Unwrap the function and look for the is_superuser property.
            return self._has_func_decorator(view, 'is_superuser')

        return False

    def get_login_required(self, view):
        return True

    def get_docstring(self, view):
        return 'Superuser required'


class UserPassesTestDecoratorProcessor(BaseDecoratorProcessor):
    """
    Process ``@user_passes_test()`` decorator.

    .. note::
        the ``@user_passes_test`` decorator does not automatically check
        that the User is not anonymous. This means they don't necessarily need
        to be authenticated for the check to pass, so this processor returns
        ``None`` (unknown) for the login_required attribute.
    """

    def can_process(self, view):
        # Some decorators use user_passes_test() internally, so we need to filter
        # them out since they are processed elsewhere.
        blacklist = (
            'is_authenticated', 'has_perms', 'is_staff', 'is_active', 'is_anonymous',
            'is_superuser',
        )

        if inspect.isclass(view):
            for func in self._get_method_decorators(view.dispatch):
                if 'user_passes_test' not in (func.__name__, func.__qualname__.split('.')[0]):
                    continue

                if not any([self._has_test_func_lambda(func, tag) for tag in blacklist]):
                    return True

        if inspect.isfunction(view) and self._has_test_func(view):
            for closure in self._get_test_func_closures(view):
                if not any([tag in closure.unbound for tag in blacklist]):
                    return True

        return False

    def get_login_required(self, view):
        return None

    def get_docstring(self, view):
        return 'Custom user test'



"""Views used for decorator testing."""
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import (
    login_required, permission_required, user_passes_test
)
from django.utils.decorators import method_decorator
from django.views.generic import View


@login_required
def login_required_view(request):
    pass


class LoginRequiredMethodDecoratorView(View):
    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


@permission_required('tests.test_perm')
def permission_required_view(request):
    pass


@permission_required(('tests.test_perm', 'tests.test_perm2'))
def permission_required_multi_view(request):
    pass


class PermissionRequiredMethodDecoratorView(View):
    @method_decorator(permission_required('tests.test_perm'))
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


@staff_member_required
def staff_member_required_view(request):
    pass


class StaffMemberRequiredMethodDecoratorView(View):
    @method_decorator(staff_member_required)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


@user_passes_test(lambda u: u.is_active)
def active_user_required_view(request):
    pass


class ActiveUserRequiredMethodDecoratorView(View):
    @method_decorator(user_passes_test(lambda u: u.is_active))
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


@user_passes_test(lambda u: u.is_anonymous)
def anonymous_user_required_view(request):
    pass


class AnonymousUserRequiredMethodDecoratorView(View):
    @method_decorator(user_passes_test(lambda u: u.is_anonymous))
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


@user_passes_test(lambda u: u.is_superuser)
def superuser_required_view(request):
    pass


class SuperUserRequiredMethodDecoratorView(View):
    @method_decorator(user_passes_test(lambda u: u.is_superuser))
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


@user_passes_test(lambda u: u.email is not None)
def user_passes_test_view(request):
    pass


class UserPassesTestMethodDecoratorView(View):
    @method_decorator(user_passes_test(lambda u: u.email is not None))
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


@login_required
@user_passes_test(lambda u: u.is_active)
@user_passes_test(lambda u: u.email is not None)
def nested_decorator_view(request):
    pass


class NestedMethodDecoratorView(View):
    @method_decorator(login_required)
    @method_decorator(user_passes_test(lambda u: u.is_active))
    @method_decorator(user_passes_test(lambda u: u.email is not None))
    def dispatch(self, request, *args, **kwargs):
        pass



"Permissions auditing for Django."

__version__ = '1.2.0'

# This can be removed once Django 3.1 and below support is dropped.
default_app_config = 'permissions_auditor.apps.PermissionsAuditorConfig'



from django.contrib import admin, messages
from django.contrib.admin import helpers
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import GroupAdmin
from django.contrib.auth.models import Group, Permission
from django.db import models
from django.db.models import Prefetch
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import path, reverse
from django.utils.html import mark_safe
from django.utils.translation import gettext_lazy as _

from permissions_auditor.core import get_views, _get_setting
from permissions_auditor.forms import AuditorAdminPermissionForm


class View(models.Model):
    """Dummy model to display views index pages in the admin."""
    class Meta:
        managed = False
        verbose_name = 'permission'
        verbose_name_plural = _('Site Views')
        app_label = 'permissions_auditor'


class ViewsIndexAdmin(admin.ModelAdmin):
    """
    Index containing all of the views found on the django site,
    and the permissions needed to access them.
    """
    form = AuditorAdminPermissionForm
    fieldsets = (
        (_('Permission Info'), {
            'fields': ('name', 'content_type', 'codename'),
        }),
        (_('Objects with this Permission'), {
            'fields': ('users', 'groups'),
        }),
    )

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        return [
            path('', self.admin_site.admin_view(self.index), name='%s_%s_changelist' % info),
            path('<str:permission>/',
                 self.admin_site.admin_view(self.permission_detail),
                 name='%s_%s_permissiondetail' % info),
        ]

    def get_object(self, request, permission, from_field=None):
        try:
            app_label, codename = permission.split('.')
            return Permission.objects.get(content_type__app_label=app_label, codename=codename)
        except (Permission.DoesNotExist, ValueError):
            return None

    def get_form(self, request, obj, change=False, **kwargs):
        defaults = {
            'users': obj.user_set.filter(is_active=True),
            'groups': obj.group_set.all()
        }
        return self.form(request.POST or defaults, instance=obj)

    def index(self, request):
        context = dict(self.admin_site.each_context(request))

        context.update({
            'views': get_views(),
            'group_by': request.GET.get('group_by', 'module')
        })
        return TemplateResponse(request, "permissions_auditor/admin/views_index.html", context)

    def permission_detail(self, request, permission, obj=None):
        try:
            obj = self.get_object(request, permission)
        except Permission.MultipleObjectsReturned:
            return self._get_obj_multiple_exist_redirect(request, permission)

        if obj is None:
            return self._get_obj_does_not_exist_redirect(request, self.model._meta, permission)

        opts = self.model._meta

        adminForm = helpers.AdminForm(
            self.get_form(request, obj),
            list(self.get_fieldsets(request, obj)),
            {},
            model_admin=self
        )
        media = self.media + adminForm.media

        if (request.method == 'POST' and
                adminForm.form.is_valid() and
                self.has_auditor_change_permission(request)):
            obj.user_set.set(adminForm.form.cleaned_data['users'])
            obj.group_set.set(adminForm.form.cleaned_data['groups'])
            return self.response_change(request, obj)

        context = {
            **self.admin_site.each_context(request),
            'adminform': adminForm,
            'errors': helpers.AdminErrorList(adminForm.form, []),
            'media': media,

            'views': get_views(),
            'permission': '{}.{}'.format(obj.content_type.app_label, obj.codename),

            'opts': opts,
            'add': False,
            'change': True,
            'is_popup': False,
            'save_as': self.save_as,
            'has_editable_inline_admin_formsets': False,
            'has_view_permission': self.has_view_permission(request, obj),
            'has_add_permission': self.has_add_permission(request, obj),
            'has_change_permission': self.has_auditor_change_permission(request),
            'has_delete_permission': self.has_delete_permission(request, obj),
            'app_label': opts.app_label,
        }

        return TemplateResponse(
            request, "permissions_auditor/admin/permission_detail.html", context
        )

    def has_view_permission(self, request, obj=None):
        return request.user.is_staff

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_module_permission(self, request):
        return self.has_view_permission(request)

    def has_auditor_change_permission(self, request):
        return request.user.has_perms(['auth.change_user', 'auth.change_group'])

    def _get_obj_multiple_exist_redirect(self, request, permission):
        """
        Create a message informing the user that multiple permissions were found
        for the specified permission string, and return to the admin index page.
        """
        msg = _('Found multiple permissions when looking up “%(permission)s”. '
                'Please ensure only a single permission exists with this name.') % {
            'permission': permission
        }
        self.message_user(request, msg, messages.WARNING)
        url = reverse('admin:index', current_app=self.admin_site.name)
        return HttpResponseRedirect(url)


class AuditorGroupAdmin(GroupAdmin):
    list_display = ['name', 'permissions_display', 'users_display']

    def permissions_display(self, obj):
        result = ''
        for perm in obj.permissions.all():
            perm_str = '{}.{}'.format(perm.content_type.app_label, perm.codename)
            url = reverse('admin:permissions_auditor_view_permissiondetail', args=(perm_str,))
            result += '<a href="{}">{}</a><br/>'.format(url, perm_str)
        return mark_safe(result)

    permissions_display.short_description = 'Permissions'

    def users_display(self, obj):
        result = ''
        for user in obj.active_users:
            url = reverse(
                'admin:{}_{}_change'.format(user._meta.app_label, user._meta.model_name),
                args=(user.pk,)
            )
            result += '<a href="{}">{}</a><br/>'.format(url, user)
        return mark_safe(result)

    users_display.short_description = 'Active Users'

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.prefetch_related(
            'permissions',
            'permissions__content_type',
            Prefetch(
                'user_set',
                queryset=get_user_model()._default_manager.filter(is_active=True),
                to_attr='active_users'
            )
        )


if _get_setting('PERMISSIONS_AUDITOR_ADMIN'):
    admin.site.register(View, ViewsIndexAdmin)

    if _get_setting('PERMISSIONS_AUDITOR_ADMIN_OVERRIDE_GROUPS'):
        admin.site.unregister(Group)
        admin.site.register(Group, AuditorGroupAdmin)



from django.conf import ImproperlyConfigured
from django.test import SimpleTestCase

from permissions_auditor import core
from permissions_auditor.core import ViewDetails
from permissions_auditor.tests.fixtures import views


class ViewParserTest(SimpleTestCase):

    def setUp(self):
        self.parser = core.ViewParser()

    def test_invalid_config(self):
        """
        Invalid processors should raise ImproperlyConfigured.
        """
        with self.settings(PERMISSIONS_AUDITOR_PROCESSORS=['invalid_parser']):
            self.assertRaises(ImproperlyConfigured, lambda: core.ViewParser())

    def test_no_parsers(self):
        with self.settings(PERMISSIONS_AUDITOR_PROCESSORS=[]):
            parser = core.ViewParser()
            results = parser.parse(views.LoginRequiredView)
            self.assertEqual(results, ([], False, ''))

    def test_parse_cbv(self):
        """
        The default configuration should be able to parse PermissionRequiredMixin.
        """
        results = self.parser.parse(views.PermissionRequiredViewDocstring)
        self.assertEqual(
            results, (['tests.test_perm'], True, 'Custom docstrings should be detected.')
        )

    def test_parse_func_view(self):
        """
        The default configuration should be able to parse @permission_required().
        """
        results = self.parser.parse(views.permission_required_view)
        self.assertEqual(
            results, (['tests.test_perm'], True, '')
        )


class GetViewsTest(SimpleTestCase):

    def setUp(self):
        self.module = 'permissions_auditor.tests.fixtures.views'
        self.views_results = [
            ViewDetails(
                module=self.module, name='BaseView',
                url='/',
                permissions=[], login_required=False, docstring=''
            ),
            ViewDetails(
                module=self.module, name='PermissionRequiredMultiView',
                url='/multi_perm_view/',
                permissions=['tests.test_perm', 'tests.test_perm2'],
                login_required=True, docstring=''
            ),
            ViewDetails(
                module=self.module, name='login_required_view',
                url='/new_style/login_required/',
                permissions=[], login_required=True, docstring=''
            ),
            ViewDetails(
                module=self.module, name='permission_required_view',
                url='/new_style/perm_required/',
                permissions=['tests.test_perm'], login_required=True, docstring=''
            ),
            ViewDetails(
                module=self.module, name='staff_member_required_view',
                url='/admin/staff_member_required/',
                permissions=[], login_required=True, docstring='Staff member required'
            ),
            ViewDetails(
                module=self.module, name='LoginRequiredView',
                url='/old_style/login_required/',
                permissions=[], login_required=True, docstring=''
            ),
            ViewDetails(
                module=self.module, name='PermissionRequiredView',
                url='/old_style/perm_required/',
                permissions=['tests.test_perm'], login_required=True, docstring=''
            ),
        ]

    def reload_blacklist(self):
        """
        The blacklist is loaded when the app starts, we need to override the values
        when we change settings.
        """
        core.NAMESPACE_BLACKLIST = tuple(core._get_blacklist('namespaces'))
        core.VIEW_BLACKLIST = tuple(core._get_blacklist('view_names'))
        core.MODULE_BLACKLIST = tuple(core._get_blacklist('modules'))

    def test_get_views_results(self):
        blacklist = {
            'namespaces': [],
            'view_names': [],
            'modules': [],
        }

        with self.settings(PERMISSIONS_AUDITOR_BLACKLIST=blacklist):
            self.reload_blacklist()
            views = core.get_views()
            self.assertSequenceEqual(views, self.views_results)

    def test_namespace_blacklist(self):
        blacklist = {
            'namespaces': ['admin'],
            'view_names': [],
            'modules': [],
        }

        with self.settings(PERMISSIONS_AUDITOR_BLACKLIST=blacklist):
            self.reload_blacklist()
            views = core.get_views()
            self.assertNotIn(views, self.views_results[3])  # staff_member_required_view

    def test_view_name_blacklist(self):
        blacklist = {
            'namespaces': [],
            'view_names': ['{}.BaseView'.format(self.module)],
            'modules': [],
        }

        with self.settings(PERMISSIONS_AUDITOR_BLACKLIST=blacklist):
            self.reload_blacklist()
            views = core.get_views()
            self.assertNotIn(views, self.views_results[0])  # BaseView

    def test_module_blacklist(self):
        blacklist = {
            'namespaces': [],
            'view_names': [],
            'modules': [self.module],
        }

        with self.settings(PERMISSIONS_AUDITOR_BLACKLIST=blacklist):
            self.reload_blacklist()
            views = core.get_views()
            self.assertEqual(views, [])



from collections import namedtuple

from django.conf import ImproperlyConfigured, settings
from django.contrib.admindocs.views import simplify_regex
from django.core.cache import cache
from django.urls.resolvers import RegexPattern, RoutePattern, URLPattern, URLResolver
from django.utils.module_loading import import_string

from . import defaults


def _get_setting(name):
    return getattr(settings, name, getattr(defaults, name))


def _get_blacklist(name):
    blacklist = _get_setting('PERMISSIONS_AUDITOR_BLACKLIST')

    # Fall back to the defauls if the user does not provide the specific
    # blacklist in their settings.
    if name not in blacklist:
        blacklist = getattr(defaults, 'PERMISSIONS_AUDITOR_BLACKLIST')

    return blacklist[name]


NAMESPACE_BLACKLIST = tuple(_get_blacklist('namespaces'))
VIEW_BLACKLIST = tuple(_get_blacklist('view_names'))
MODULE_BLACKLIST = tuple(_get_blacklist('modules'))

ViewDetails = namedtuple('ViewDetails', [
    'module', 'name', 'url', 'permissions', 'login_required', 'docstring'
])


class ViewParser:

    def __init__(self):
        self.load_processors()

    def load_processors(self):
        self._processors = []

        for processor_path in _get_setting('PERMISSIONS_AUDITOR_PROCESSORS'):

            try:
                processor = import_string(processor_path)
                self._processors.append(processor())
            except (ImportError, TypeError) as ex:
                raise ImproperlyConfigured(
                    '{} is not a valid permissions processor.'.format(processor_path)
                ) from ex

    def parse(self, view):
        """
        Process a view.

        Returns a tuple containing:
        permissions (list), login_required (boolean or None), docstring (str)
        """
        permissions = []
        login_required = False
        docstrings = []

        for processor in self._processors:
            if processor.can_process(view):
                permissions.extend(processor.get_permission_required(view))

                login_required_result = processor.get_login_required(view)
                if login_required_result is None and not login_required:
                    login_required = login_required_result
                else:
                    login_required = login_required_result or login_required

                docstrings.append(processor.get_docstring(view))

        return (
            sorted(set(permissions)),
            login_required,
            '\n'.join(list(set(filter(None, docstrings))))
        )


def _get_views(urlpatterns=None, base_url=''):
    """
    Recursively fetch all views in the specified urlpatterns.

    If urlpatterns is not specified, uses the `PERMISSIONS_AUDITOR_ROOT_URLCONF`
    setting, which by default is the value of `ROOT_URLCONF` in your project settings.

    Returns a list of `View` namedtuples.
    """
    views = []

    if urlpatterns is None:
        root_urlconf = import_string(_get_setting('PERMISSIONS_AUDITOR_ROOT_URLCONF'))
        urlpatterns = root_urlconf.urlpatterns

    parser = ViewParser()

    for pattern in urlpatterns:
        if isinstance(pattern, RoutePattern) or isinstance(pattern, URLResolver):

            if pattern.namespace in NAMESPACE_BLACKLIST:
                continue

            # Recursively fetch patterns
            views.extend(_get_views(pattern.url_patterns, base_url + str(pattern.pattern)))

        elif isinstance(pattern, URLPattern) or isinstance(pattern, RegexPattern):
            view = pattern.callback

            # If this is a CBV, use the actual class instead of the as_view() classmethod.
            view = getattr(view, 'view_class', view)
            view = getattr(view, 'cls', view)

            full_view_path = '{}.{}'.format(view.__module__, view.__name__)
            if full_view_path in VIEW_BLACKLIST or view.__module__ in MODULE_BLACKLIST:
                continue

            permissions, login_required, docstring = parser.parse(view)

            views.append(ViewDetails._make([
                view.__module__,
                view.__name__,
                simplify_regex(base_url + str(pattern.pattern)),
                permissions,
                login_required,
                docstring
            ]))

    return views


def get_views(urlpatterns=None, base_url=''):
    """
    Wrapper for caching _get_views().
    """
    cache_key = _get_setting('PERMISSIONS_AUDITOR_CACHE_KEY')
    cache_base_url_key = cache_key + '_BASE_URL'
    cache_timeout = _get_setting('PERMISSIONS_AUDITOR_CACHE_TIMEOUT')

    cache_content = cache.get_many([cache_key, cache_base_url_key])
    views = cache_content.get(cache_key, None)
    cached_base_url = cache_content.get(cache_base_url_key, None)

    if views is None or cached_base_url != base_url:
        views = _get_views(urlpatterns, base_url)
        cache.set_many({
            cache_key: views,
            cache_base_url_key: base_url
        }, timeout=cache_timeout)

    return views



from django.apps import AppConfig
from django.core.cache import cache
from django.utils.translation import gettext_lazy as _

from permissions_auditor.core import _get_setting


class PermissionsAuditorConfig(AppConfig):
    name = 'permissions_auditor'
    verbose_name = _('Permissions Auditor')

    default_auto_field = 'django.db.models.BigAutoField'

    def ready(self):
        # Delete the cached views list on application reload.
        cache.delete_many([
            _get_setting('PERMISSIONS_AUDITOR_CACHE_KEY'),
            _get_setting('PERMISSIONS_AUDITOR_CACHE_KEY') + '_BASE_URL',
        ])



from django.contrib.auth.views import PasswordChangeView

from permissions_auditor.processors import auth_decorators
from permissions_auditor.tests.base import ProcessorTestCase
from permissions_auditor.tests.fixtures import (
    base_views, decorator_views as views, mixin_views
)


class DecoratorProcessorTestCaseMixin:
    """
    Decorator processors should not be able to process class based views
    that do not use ``@method_decorator``.
    """

    def test_cannot_process_non_method_decorator_cbvs(self):
        self.assertCannotProcess([
            base_views.BaseView, base_views.base_view,
            mixin_views.LoginRequiredView,
            mixin_views.PermissionRequiredView,
            mixin_views.UserPassesTestView
        ])


class TestLoginRequiredDecoratorProcessor(DecoratorProcessorTestCaseMixin, ProcessorTestCase):

    def setUp(self):
        self.processor = auth_decorators.LoginRequiredDecoratorProcessor()
        self.expected_results = {'login_required': True}

    def test_cannot_process(self):
        self.assertCannotProcess([
            views.permission_required_view, views.PermissionRequiredMethodDecoratorView,
            views.staff_member_required_view, views.StaffMemberRequiredMethodDecoratorView,
            views.active_user_required_view, views.ActiveUserRequiredMethodDecoratorView,
            views.anonymous_user_required_view, views.AnonymousUserRequiredMethodDecoratorView,
            views.superuser_required_view, views.SuperUserRequiredMethodDecoratorView,
            views.user_passes_test_view, views.UserPassesTestMethodDecoratorView
        ])

    def test_django_password_change_view(self):
        self.assertCanProcessView(PasswordChangeView, **self.expected_results)

    def test_login_required_view(self):
        self.assertCanProcessView(views.login_required_view, **self.expected_results)

    def test_cb_loginrequired_decorator_view(self):
        self.assertCanProcessView(views.LoginRequiredMethodDecoratorView, **self.expected_results)

    def test_nested_decorator_view(self):
        self.assertCanProcessView(views.nested_decorator_view, **self.expected_results)

    def test_nested_decorator_cbv(self):
        self.assertCanProcessView(views.NestedMethodDecoratorView, **self.expected_results)


class TestPermissionRequiredDecoratorProcessor(DecoratorProcessorTestCaseMixin, ProcessorTestCase):

    def setUp(self):
        self.processor = auth_decorators.PermissionRequiredDecoratorProcessor()
        self.expected_results = {'permissions': ['tests.test_perm'], 'login_required': True}

    def test_cannot_process(self):
        self.assertCannotProcess([
            views.login_required_view, views.LoginRequiredMethodDecoratorView,
            views.staff_member_required_view, views.StaffMemberRequiredMethodDecoratorView,
            views.active_user_required_view, views.ActiveUserRequiredMethodDecoratorView,
            views.anonymous_user_required_view, views.AnonymousUserRequiredMethodDecoratorView,
            views.superuser_required_view, views.SuperUserRequiredMethodDecoratorView,
            views.user_passes_test_view, views.UserPassesTestMethodDecoratorView,
            views.nested_decorator_view, views.NestedMethodDecoratorView
        ])

    def test_permission_required_view(self):
        self.assertCanProcessView(
            views.PermissionRequiredMethodDecoratorView, **self.expected_results
        )

    def test_permission_required_cbv(self):
        self.assertCanProcessView(views.permission_required_view, **self.expected_results)

    def test_permission_required_multi_view(self):
        """Multiple permissions passed to @permission_required should be retrieved."""
        self.assertCanProcessView(
            views.permission_required_multi_view,
            permissions=['tests.test_perm', 'tests.test_perm2'], login_required=True,
        )


class StaffMemberRequiredDecoratorProcessor(DecoratorProcessorTestCaseMixin, ProcessorTestCase):

    def setUp(self):
        self.processor = auth_decorators.StaffMemberRequiredDecoratorProcessor()
        self.expected_results = {'login_required': True, 'docstring': 'Staff member required'}

    def test_cannot_process(self):
        self.assertCannotProcess([
            views.login_required_view, views.LoginRequiredMethodDecoratorView,
            views.permission_required_view, views.PermissionRequiredMethodDecoratorView,
            views.active_user_required_view, views.ActiveUserRequiredMethodDecoratorView,
            views.anonymous_user_required_view, views.AnonymousUserRequiredMethodDecoratorView,
            views.superuser_required_view, views.SuperUserRequiredMethodDecoratorView,
            views.user_passes_test_view, views.UserPassesTestMethodDecoratorView,
            views.nested_decorator_view, views.NestedMethodDecoratorView
        ])

    def test_staff_member_required_view(self):
        self.assertCanProcessView(views.staff_member_required_view, **self.expected_results)

    def test_staff_member_required_cb_view(self):
        self.assertCanProcessView(
            views.StaffMemberRequiredMethodDecoratorView, **self.expected_results
        )


class ActiveUserRequiredDecoratorProcessor(DecoratorProcessorTestCaseMixin, ProcessorTestCase):

    def setUp(self):
        self.processor = auth_decorators.ActiveUserRequiredDecoratorProcessor()
        self.expected_results = {'login_required': True, 'docstring': 'Active user required'}

    def test_cannot_process(self):
        self.assertCannotProcess([
            views.login_required_view, views.LoginRequiredMethodDecoratorView,
            views.permission_required_view, views.PermissionRequiredMethodDecoratorView,
            views.anonymous_user_required_view, views.AnonymousUserRequiredMethodDecoratorView,
            views.superuser_required_view, views.SuperUserRequiredMethodDecoratorView,
            views.user_passes_test_view, views.UserPassesTestMethodDecoratorView
        ])

    def test_staff_member_required_view(self):
        # Internally, ``staff_member_required`` runs this lamda:
        # lambda u: u.is_active and u.is_staff
        self.assertCanProcessView(views.staff_member_required_view, **self.expected_results)

    def test_staff_member_required_cbv(self):
        # Quirk: when ``staf_member_required`` is wrapped with ``method_decorator``,
        # the lamda expression can no longer be found in the function closures,
        # so this should return false.
        self.assertCannotProcess([views.StaffMemberRequiredMethodDecoratorView])

    def test_active_user_required_view(self):
        self.assertCanProcessView(views.active_user_required_view, **self.expected_results)

    def test_active_user_required_cbv(self):
        self.assertCanProcessView(
            views.ActiveUserRequiredMethodDecoratorView, **self.expected_results
        )

    def test_nested_decorator_view(self):
        self.assertCanProcessView(views.nested_decorator_view, **self.expected_results)

    def test_nested_decorator_cbv(self):
        self.assertCanProcessView(views.NestedMethodDecoratorView, **self.expected_results)


class AnonymousUserRequiredDecoratorProcessor(DecoratorProcessorTestCaseMixin, ProcessorTestCase):

    def setUp(self):
        self.processor = auth_decorators.AnonymousUserRequiredDecoratorProcessor()
        self.expected_results = {'login_required': False, 'docstring': 'Anonymous user required'}

    def test_cannot_process(self):
        self.assertCannotProcess([
            views.login_required_view, views.LoginRequiredMethodDecoratorView,
            views.permission_required_view, views.PermissionRequiredMethodDecoratorView,
            views.staff_member_required_view, views.StaffMemberRequiredMethodDecoratorView,
            views.active_user_required_view, views.ActiveUserRequiredMethodDecoratorView,
            views.superuser_required_view, views.SuperUserRequiredMethodDecoratorView,
            views.user_passes_test_view, views.UserPassesTestMethodDecoratorView,
            views.nested_decorator_view, views.NestedMethodDecoratorView
        ])

    def test_anonymous_user_required_view(self):
        self.assertCanProcessView(views.anonymous_user_required_view, **self.expected_results)

    def test_anonymous_user_required_cbv(self):
        self.assertCanProcessView(
            views.AnonymousUserRequiredMethodDecoratorView, **self.expected_results
        )


class SuperUserRequiredDecoratorProcessor(DecoratorProcessorTestCaseMixin, ProcessorTestCase):

    def setUp(self):
        self.processor = auth_decorators.SuperUserRequiredDecoratorProcessor()
        self.expected_results = {'login_required': True, 'docstring': 'Superuser required'}

    def test_cannot_process(self):
        self.assertCannotProcess([
            views.login_required_view, views.LoginRequiredMethodDecoratorView,
            views.permission_required_view, views.PermissionRequiredMethodDecoratorView,
            views.staff_member_required_view, views.StaffMemberRequiredMethodDecoratorView,
            views.active_user_required_view, views.ActiveUserRequiredMethodDecoratorView,
            views.anonymous_user_required_view, views.AnonymousUserRequiredMethodDecoratorView,
            views.user_passes_test_view, views.UserPassesTestMethodDecoratorView,
            views.nested_decorator_view, views.NestedMethodDecoratorView
        ])

    def test_superuser_required_view(self):
        self.assertCanProcessView(views.superuser_required_view, **self.expected_results)

    def test_superuser_required_cbv(self):
        self.assertCanProcessView(
            views.SuperUserRequiredMethodDecoratorView, **self.expected_results
        )


class UserPassesTestDecoratorProcessor(DecoratorProcessorTestCaseMixin, ProcessorTestCase):

    def setUp(self):
        self.processor = auth_decorators.UserPassesTestDecoratorProcessor()
        self.expected_results = {'login_required': None, 'docstring': 'Custom user test'}

    def test_cannot_process(self):
        self.assertCannotProcess([
            views.login_required_view, views.LoginRequiredMethodDecoratorView,
            views.permission_required_view, views.PermissionRequiredMethodDecoratorView,
            views.staff_member_required_view, views.StaffMemberRequiredMethodDecoratorView,
            views.active_user_required_view, views.ActiveUserRequiredMethodDecoratorView,
            views.anonymous_user_required_view, views.AnonymousUserRequiredMethodDecoratorView,
            views.superuser_required_view, views.SuperUserRequiredMethodDecoratorView,
        ])

    def test_user_passes_test_view(self):
        self.assertCanProcessView(views.user_passes_test_view, **self.expected_results)

    def test_user_passes_test_cb_view(self):
        self.assertCanProcessView(views.UserPassesTestMethodDecoratorView, **self.expected_results)

    def test_nested_decorator_view(self):
        self.assertCanProcessView(views.nested_decorator_view, **self.expected_results)

    def test_nested_decorator_cbv(self):
        self.assertCanProcessView(views.NestedMethodDecoratorView, **self.expected_results)




from django import VERSION as DJANGO_VERSION

if DJANGO_VERSION < (4, 0):
    # This can be removed once Django 3.2 and below support is dropped.
    from django.conf.urls import url, include as old_include

from django.urls import path, include, re_path

from . import views

new_style_urls = ([
    path('login_required/', views.login_required_view),
    re_path(r'^perm_required/$', views.permission_required_view),
], 'new_style_urls')

admin_namespace = ([
    path('staff_member_required/', views.staff_member_required_view),
], 'admin')

urlpatterns = [
    path('', views.BaseView.as_view()),
    path('multi_perm_view/', views.PermissionRequiredMultiView.as_view()),

    path('new_style/', include(new_style_urls)),
    path('admin/', include(admin_namespace)),
]

if DJANGO_VERSION < (4, 0):
    # This can be removed once Django 3.2 and below support is dropped.
    old_style_urls = ([
        url(r'^login_required/$', views.LoginRequiredView.as_view()),
        url(r'^perm_required/$', views.PermissionRequiredView.as_view()),
    ], 'old_style_urls')

    urlpatterns += [url('old_style/', old_include(old_style_urls))]
else:
    old_style_urls = ([
        re_path(r'^login_required/$', views.LoginRequiredView.as_view()),
        re_path(r'^perm_required/$', views.PermissionRequiredView.as_view()),
    ], 'old_style_urls')

    urlpatterns += [path('old_style/', include(old_style_urls))]



"""Processors for django auth mixins."""

import inspect

from django.conf import ImproperlyConfigured

from .base import BaseFilteredMixinProcessor


class PermissionRequiredMixinProcessor(BaseFilteredMixinProcessor):
    """
    Processes views that inherit from
    ``django.contrib.auth.mixins.PermissionRequiredMixin``.

    .. hint::
        If the ``has_permission()`` function is overridden, any docstrings on that
        function will be displayed in the additional info column.
    """

    class_filter = 'django.contrib.auth.mixins.PermissionRequiredMixin'

    def get_permission_required(self, view):
        try:
            return view().get_permission_required()
        except ImproperlyConfigured:
            return []

    def get_login_required(self, view):
        return True

    def get_docstring(self, view):
        docstring = None

        # Check if has_permission has been overriden.
        if 'has_permission' in view.__dict__:
            docstring = inspect.getdoc(view.has_permission)

            if docstring is None or docstring.startswith('Override this method'):
                docstring = 'Custom (no docstring found)'

        return docstring


class LoginRequiredMixinProcessor(BaseFilteredMixinProcessor):
    """
    Processes views that inherit from
    ``django.contrib.auth.mixins.LoginRequiredMixin``.
    """

    class_filter = 'django.contrib.auth.mixins.LoginRequiredMixin'

    def get_login_required(self, view):
        return True


class UserPassesTestMixinProcessor(BaseFilteredMixinProcessor):
    """
    Processes views that inherit from
    ``django.contrib.auth.mixins.UserPassesTestMixin``.

    .. hint::
        If the function returned by ``get_test_func()`` is overridden, any docstrings
        on that function will be displayed in the additional info column.

    .. note::
        UserPassesTestMixinProcessor does not automatically check
        that the User is not anonymous. This means they don't necessarily need
        to be authenticated for the check to pass, so this processor returns
        ``None`` (unknown) for the login_required attribute.
    """

    class_filter = 'django.contrib.auth.mixins.UserPassesTestMixin'

    def get_login_required(self, view):
        return None

    def get_docstring(self, view):
        docstring = inspect.getdoc(view().get_test_func())
        if docstring is None:
            docstring = 'Custom (no docstring found)'
        return docstring



from unittest.mock import patch

from django.conf import ImproperlyConfigured

from permissions_auditor.processors import base
from permissions_auditor.tests.base import ProcessorTestCase
from permissions_auditor.tests.fixtures import views


class BaseProcessorTest(ProcessorTestCase):
    def setUp(self):
        self.processor = base.BaseProcessor()

    def test_process(self):
        self.assertCanProcessView(views.BaseView)


class BaseFuncViewProcessorTest(ProcessorTestCase):
    def setUp(self):
        self.processor = base.BaseFuncViewProcessor()

    def test_can_process_class(self):
        self.assertCannotProcess([views.BaseView])

    def test_can_process_function(self):
        self.assertCanProcessView(views.base_view)


class BaseCBVProcessorTest(ProcessorTestCase):
    def setUp(self):
        self.processor = base.BaseCBVProcessor()

    def test_can_process_class(self):
        self.assertCanProcessView(views.BaseView)

    def test_can_process_function(self):
        self.assertCannotProcess([views.base_view])


class BaseFilteredMixinProcessorTest(ProcessorTestCase):
    def setUp(self):
        self.processor = base.BaseFilteredMixinProcessor()

    def test_no_class_filter_raises_exception(self):
        with self.assertRaises(ImproperlyConfigured):
            self.assertCannotProcess([views.PermissionRequiredView])

    def test_can_process_filtered_class(self):
        self.processor.class_filter = 'django.contrib.auth.mixins.PermissionRequiredMixin'
        self.assertCanProcessView(views.PermissionRequiredView)
        self.assertCannotProcess([views.LoginRequiredView])

    def test_can_process_multiple_filtered_classses(self):
        self.processor.class_filter = (
            'django.contrib.auth.mixins.PermissionRequiredMixin',
            'django.contrib.auth.mixins.LoginRequiredMixin',
        )
        self.assertCanProcessView(views.PermissionRequiredView)
        self.assertCanProcessView(views.LoginRequiredView)

    def test_can_process_overriden_filtered_class(self):
        """
        can_process() should use use get_class_filter() if it is overridden.
        """
        self.processor.class_filter = None

        def filter_func():
            return ('django.contrib.auth.mixins.LoginRequiredMixin',)

        with patch.object(self.processor, 'get_class_filter', filter_func):
            self.assertCannotProcess([views.PermissionRequiredView])
            self.assertCanProcessView(views.LoginRequiredView)



from django.test import SimpleTestCase


class ProcessorTestCase(SimpleTestCase):
    processor = None

    def assertCannotProcess(self, views):
        for view in views:
            self.assertFalse(self.processor.can_process(view))

    def assertCanProcessView(self, view, permissions=[], login_required=False, docstring=None):
        self.assertTrue(self.processor.can_process(view))
        self.assertCountEqual(self.processor.get_permission_required(view), permissions)
        self.assertEqual(self.processor.get_login_required(view), login_required)
        self.assertEqual(self.processor.get_docstring(view), docstring)



import csv
import json

from django.core.management.base import BaseCommand

from permissions_auditor.core import get_views


class Command(BaseCommand):
    help = 'Dumps all detected view permissions to the specified output format.'

    def add_arguments(self, parser):
        parser.add_argument(
            "--format",
            default="json",
            help="Specifies the output serialization format for permissions. Options: csv, json",
        )
        parser.add_argument(
            "-o", "--output", help="Specifies file to which the output is written."
        )

    def handle(self, *args, **options):
        format = options["format"]
        output = options["output"]
        views = get_views()

        if format == 'csv':
            if output:
                with open(output, 'w', newline='') as file:
                    writer = csv.writer(file, dialect='excel')
                    self._write_csv(views, writer)
            else:
                writer = csv.writer(self.stdout)
                self._write_csv(views, writer)

        elif format == 'json':
            data = [v._asdict() for v in views]

            if output:
                with open(output, 'w') as file:
                    json.dump(data, file, indent=4)
            else:
                self.stdout.write(json.dumps(data))
        else:
            raise NotImplementedError('Output format `{}` is not implemented.'.format(format))

    def _write_csv(self, views, writer):
        # Header
        writer.writerow(['module', 'name', 'url', 'permissions', 'login_required', 'docstring'])

        for view in views:
            writer.writerow(view)
